/*
 * This file is part of firestorm NIDS
 * Copyright (c) 2002 Gianni Tedesco
 *
 * 	The Boyer Moore implementation was pretty much entirely
 * 	taken from snorts mstring.c. The original copyright notice
 * 	of which is reproduced below. That covers the functions
 *	bm_skip(), bm_shift() and str_match(). It hardly matters
 *	as both firestorm and snort are GPL anyway :)
 */

/*
** Copyright (C) 1998-2002 Martin Roesch <roesch@sourcefire.com>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/


#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <netinet/in.h>

#include <packet.h>
#include <matcher.h>
#include <plugin.h>

#define DSM_EQ	0
#define DSM_LT	1
#define DSM_GT	2

struct ds_priv {
	unsigned long	dsize;
};

struct content_priv {
	char 		*string;
	unsigned int	len;

	/* Boyer moore skip/shift tables */
	int		skip[256];
	int		*shift;
};

/* dsize matcher: match packet payload size */
int ds_match_eq(struct packet *p, void *priv, unsigned int l, int n)
{
	struct ds_priv *dsp=priv;
	unsigned int dsize;
	dsize=p->end - p->layer[p->llen-1].h.raw;
	return n ^ (dsize==dsp->dsize);
}

int ds_match_gt(struct packet *p, void *priv, unsigned int l, int n)
{
	struct ds_priv *dsp=priv;
	unsigned int dsize;
	dsize=p->end - p->layer[p->llen-1].h.raw;
	return n ^ (dsize>dsp->dsize);
}

int ds_match_lt(struct packet *p, void *priv, unsigned int l, int n)
{
	struct ds_priv *dsp=priv;
	unsigned int dsize;
	dsize=p->end - p->layer[p->llen-1].h.raw;
	return n ^ (dsize<dsp->dsize);
}

int ds_compare(void *p1, void *p2)
{
	struct ds_priv *dsp1=p1;
	struct ds_priv *dsp2=p2;

	return ( (dsp1->dsize==dsp2->dsize) );
}

proc_match_match ds_validate(char *args, void **priv)
{
	struct ds_priv *dsp;
	proc_match_match ret=ds_match_eq;
	unsigned int num;
	char *val;

	if ( !args ) return NULL;
	
	for(val=args; *val; val++) {
		if ( *val>='0' && *val<='9' ) {
			break;	
		}
		if ( *val=='<' ) {
			ret=ds_match_lt;
		}else if ( *val=='>' ){
			ret=ds_match_gt;
		}
	}
	
	if ( (num=strtoul(val, NULL, 0))==ULONG_MAX &&
		errno==ERANGE ) return NULL;

	if ( !(dsp=malloc(sizeof(*dsp))) )
		return NULL;

	dsp->dsize=num;
	*priv=dsp;

	return ret;
}

/* Content matcher (generic string matching) */
void bm_skip(char *x, int m, int *skip) 
{
	int i;
 
	for (i = 0; i < 256; ++i)
		skip[i] = m+1;
	
	while(m)
		skip[(unsigned char)*x++]=m--;
}
 
void bm_shift(char *ptrn, int plen, int *shift)
{
	int *sptr=shift+plen-1;
	char *pptr=ptrn+plen-1;
	char c;

	c=ptrn[plen-1];
	*sptr=1;

	while(sptr-- != shift) {
		char *p1=ptrn+plen-2, *p2, *p3;

		do {
			while(p1>=ptrn && *p1-- !=c );

			p2=ptrn+plen-2;
			p3=p1;

			while(p3>=ptrn &&
				*p3-- == *p2-- &&
				p2 >= pptr);
		}

		while(p3>=ptrn && 
			*p3-- == *p2-- && 
			p2 >= pptr);

		*sptr=shift+plen-sptr+p2-p3;

		pptr--;
	}
}
 
int str_match(struct packet *p, void *priv, unsigned int l, int n)
{
	struct content_priv *sp=priv;
	unsigned int hlen;
	char *hs;
	int b_idx;

	hs=p->layer[p->llen-1].h.raw;
	hlen=(unsigned int)(p->end - (void *)hs);

	b_idx=sp->len;
	while(b_idx <= hlen ) {
		int skip_stride,shift_stride, p_idx=sp->len;
		
		while(hs[--b_idx]==sp->string[--p_idx]) {
			if (b_idx<0) return n^0;
			if (p_idx==0) return n^1;
		}

		skip_stride=sp->skip[(unsigned char)hs[b_idx]];
		shift_stride=sp->shift[p_idx];

		b_idx += (skip_stride > shift_stride) ? skip_stride : shift_stride;
	}
	
	return n^0;
}

 
proc_match_match str_validate(char *args, void **priv)
{
	char *t;
	int state=0;
	char bin[3]={0,0,0};
	unsigned int i=0, j=0;
	struct content_priv *p;

	if ( !args ) return NULL;

	if ( !(j=strlen(args)) ) return NULL;
	
	if ( !(p=calloc(1, sizeof(*p))) ) {
		return NULL;
	}

	/* Finished strings can be no bigger than this */
	if ( !(p->string=malloc(j)) ) {
		free(p);
		return NULL;
	}p->len=j;

	for(t=args,j=0; *t; t++) {
		if ( state==0 ) { /* Normal string */
			if ( *t=='|' ) {
				state=1;
			}else {
				p->string[j++]=*t;
			}
		}else if ( state==1 ) { /* Binary data */
			if ( *t=='|' ) {
				state=0;
				continue;
			}
			
			if ( *t==' ' || *t=='\t' ) {
				continue;
			}

			bin[i++]=*t;

			if ( i==2 ) {
				unsigned long int ul;
				i=0;
				ul=strtoul(bin, NULL, 16);

				/* Impossible to be too big cos
				 * 16^2 = 0xff */
				p->string[j++]=(char )ul&0xff;
			}
		}
	}

	if ( i ) {
		printf("string is n and a half bytes!\n");
		free(p->string);
		free(p);
		return NULL;
	}

	/* Shrink the string if there was any
	 * binary data inside of it */
	if ( j!=p->len ) {
		realloc(p->string, j); /* can only be smaller */
		p->len=j;
	}

	/* Perform boyer-moore preprocessing */
	if ( !(p->shift=(int *)malloc(p->len*sizeof(int))) ) {
		free(p->string);
		free(p);
		return NULL;
	}

	bm_shift(p->string, p->len, p->shift);
	bm_skip(p->string, p->len, p->skip);

	*priv=p;
	return str_match;
}

int str_compare(void *p1, void *p2)
{
	return 0;
}

void str_cleanup(void *priv) {
	struct content_priv *p=priv;

	if ( p ) {
		if ( p->string ) free(p->string);
		free(p);
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	int ok=0;
	proc_matcher_add matcher_add;
	
	plugin_check(in, out);
	
	PLUGIN_ID("match.std", "Generic matching routines");
	PLUGIN_VERSION(2, 0);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");

	if ( !(matcher_add=in->import("matcher.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=matcher_add("dsize",
		ds_validate,
		ds_compare,
		in->import("template.ptr"));
	ok+=matcher_add("content",
		str_validate,
		str_compare,
		str_cleanup);
	ok+=matcher_add("uricontent",
		str_validate,
		str_compare,
		str_cleanup);
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	return PLUGIN_ERR_OK;
}
