#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <netinet/in.h>
#include <packet.h>
#include <preproc.h>
#include <decode.h>
#include <plugin.h>

/* 
 * IP Defragmentation for firestorm
 * ================================
 * 
 * This code is a reworking of ip_fragment.c from Linux 2.4.18,
 * it should be relatively straight forward to understand. This
 * is a first cut so it still might not be perfect.
 *
 * Should cope with:
 *  Overlapping fragments
 *  Oversized fragments
 *  Out of order fragments
 *
 * TODO:
 *  Calculate checksum on reassembled packets
 *  Handle tunables
 *  evictor optimisations (global lru/cache tail)
 *  Timeout packets (worth bothering?)
 *  Coalesce fragments in incomplete packets (??)
 */

#if 1
#define dprintf(x, ...) 
#define __INLINE__ static inline
#else
#define dprintf printf
#define __INLINE__
#endif

/* Statistics */
unsigned int ipfrag_errs=0;
unsigned int ipfrag_reassembled=0;

/* Memory usage, high and low water marks */
unsigned int ipfrag_mem=0;
unsigned int ipfrag_mem_hi=2048*1024;
unsigned int ipfrag_mem_lo=3072*1024;

/* Stuff we import */
proc_preproc_activate preproc_activate;
proc_preproc_dispatch preproc_dispatch;
struct proto *ipproto;

/* Mask to determine fragments */
unsigned short ipfmask=__constant_htons(IP_MF|IP_OFFMASK);

/* Reassembly */
struct packet ipfr;
char *reasm_buf=NULL;
unsigned int reasm_len=0;

/* Keeps each individual fragment */
struct ipfrag {
	struct ipfrag		*next;
	unsigned int		len;
	unsigned int		offset;
	void			*data;
	unsigned int		free;
	void			*fdata; /* Data to free */
};

struct ipq {
	struct ipq	*next;
	struct ipq	**pprev;

	/* Identify the packet */
	u_int32_t	saddr;
	u_int32_t	daddr;
	u_int16_t	id;
	u_int8_t	protocol;

	/* Linked list of fragments */
	struct ipfrag	*fragments;
	unsigned int	num_frags;

#define FIRST_IN	2
#define LAST_IN		1
	int		last_in;

	/* Total size of all the fragments we have */
	int		meat;

	/* Total length of full packet */
	int		len;

	/* Stuff we need for reassembly */
	struct cap_s	*cap;
	u_int32_t	flags;
};

#define IPHASH 128 /* Must be a power of two */
struct ipq *ipq_hash[IPHASH]; /* IP fragment hash table */

/* Hash function for ipq_hash lookup */
__INLINE__ unsigned int ipq_hashfn(u_int16_t id,
	u_int32_t saddr,
	u_int32_t daddr,
	u_int8_t proto){
	unsigned int h=saddr^daddr;

	h ^= (h>>16)^id;
	h ^= (h>>8)^proto;
	return h & (IPHASH-1);
}

void ipq_kill(struct ipq *qp)
{
	struct ipfrag *foo, *bar;

	/* Unlink from the list */
	if ( qp->next )
		qp->next->pprev=qp->pprev;
	*qp->pprev=qp->next;

	/* Free the fragments and descriptors */
	for(foo=qp->fragments; foo;) {
		bar=foo;
		foo=foo->next;
		if ( bar->free ) free(bar->fdata);
		free(bar);
		ipfrag_mem-=sizeof(struct ipfrag);
	}

	/* Free the ipq itself */
	free(qp);
	ipfrag_mem-=sizeof(struct ipq);
}

/* Reassemble a complete set of fragments, decode the
 * new packet, and send it back through the preprocessor
 * list, we won't touch it next time round */
__INLINE__ void ipfrag_reassemble(struct ipq *qp)
{
	struct ipfrag *f;
	struct pkt_iphdr *iph;
	unsigned int olen=qp->len;
	unsigned int len=0;
	unsigned int hlen=0;
	char *buf;

	/* Kill oversize packets */
	if ( olen > 65535 ) {
		ipq_kill(qp);
		ipfrag_errs++;
		return;
	}

	iph=qp->fragments->data;
	hlen=iph->ihl<<2;
	qp->fragments->len+=hlen;
	olen+=hlen;

	/* Allocate the frankenpacket buffer */
	if ( reasm_buf ) {
		char *tmp;
		
		/* Expand the buffer if it's too small */
		if ( olen < reasm_len ) {
			if ( !(tmp=realloc(reasm_buf, olen)) ) {
				olen=reasm_len;
			}else{
				reasm_buf=tmp;
				reasm_len=qp->len;
			}
		}
	}else{
		if ( !(reasm_buf=malloc(olen)) ) goto reasm_err;
		reasm_len=olen;
	}
	
	/* Copy all the fragments in to the new buffer */
	dprintf("Reassemble: %u bytes\n", olen);
	buf=reasm_buf;
	for(f=qp->fragments; f; f=f->next) {
		dprintf(" * %u bytes @ %u\n", f->len, f->offset);
		memcpy(buf, f->data, f->len);
		buf+=f->len;
		len+=f->len;
		if ( len >= olen ) break;
	}
	dprintf("\n");

	/* Build the packet */
	ipfr.len=len;
	ipfr.caplen=len;
	ipfr.llen=1;
	ipfr.layer[0].proto=ipproto;
	ipfr.layer[0].h.raw=reasm_buf;
	ipfr.base=reasm_buf;
	ipfr.capture=qp->cap;
	ipfr.flags=qp->flags | FP_CLONE;
	
	/* Kill the queue entry */
	ipq_kill(qp);

	/* Fixup the IP header */
	ipfr.layer[0].h.ip->frag_off=0;
	ipfr.layer[0].h.ip->csum=0;
	ipfr.layer[0].h.ip->tot_len=ntohs(len);

	/* Inject the packet back in to the flow */
	ipfrag_reassembled++;
	if ( ipfr.layer[0].proto->decode(&ipfr) ) {
		preproc_dispatch(&ipfr);
	}
	
	return;

reasm_err:
	ipfrag_errs++;
	printf("ipfrag: reassemble error!\n");
	return;
}

__INLINE__ struct ipq *ip_frag_create(unsigned int hash, struct pkt_iphdr *iph)
{ 
	struct ipq *q;
	
	if ( !(q=calloc(1, sizeof(struct ipq))) ){
		ipfrag_errs++;
		return NULL;
	}
	ipfrag_mem+=sizeof(struct ipq);

	q->id=iph->id;
	q->saddr=iph->saddr;
	q->daddr=iph->daddr;
	q->protocol=iph->protocol;

	if ( (q->next=ipq_hash[hash]) ) {
		q->next->pprev=&q->next;
	}
	ipq_hash[hash]=q;
	q->pprev=&ipq_hash[hash];

	return q;
}

/* Find (or create) the ipq for this IP fragment */
__INLINE__ struct ipq *ip_find(struct pkt_iphdr *iph, unsigned int *hash)
{
	struct ipq *qp;

	*hash=ipq_hashfn(iph->id, iph->saddr,
		iph->daddr, iph->protocol);

	for(qp=ipq_hash[*hash]; qp; qp=qp->next) {
		if ( qp->id == iph->id &&
			qp->saddr == iph->saddr &&
			qp->daddr == iph->daddr &&
			qp->protocol == iph->protocol ) {
			return qp;
		}
	}

	return ip_frag_create(*hash, iph);
}

void ip_evictor(struct ipq *cq)
{
	int i, progress;
	struct ipq *qp;
	
	dprintf("Running the ipfrag evictor!\n");
	do {
		if ( ipfrag_mem <= ipfrag_mem_lo ) return;

		progress=0;

		for(i=0; i<IPHASH; i++) {
			if ( !(qp=ipq_hash[i]) ) continue;

			while(qp->next)
				qp = qp->next;

			if ( qp != cq ) {
				ipfrag_errs++;
				progress=1;
				ipq_kill(qp);
			}
		}
	}while(progress);
}

__INLINE__ void ipfrag_queue(
	unsigned int hash,
	struct ipq *qp, 
	struct packet *pkt, 
	struct pkt_iphdr *iph)
{
	struct ipfrag *prev, *next, *me;
	int flags, offset;
	int ihl, end, len;
	int chop=0;

	/* Remove from queue */
	if ( qp->next )
		qp->next->pprev=qp->pprev;
	*qp->pprev=qp->next;

	/* Prepend to queue */
	if ( (qp->next=ipq_hash[hash]) )
		qp->next->pprev=&qp->next;
	ipq_hash[hash]=qp;
	qp->pprev=&ipq_hash[hash];

	/* Kill off LRU ipqs */
	if ( ipfrag_mem > ipfrag_mem_hi ) {
		ip_evictor(qp);
	}

	offset=ntohs(iph->frag_off);
	flags=offset & ~IP_OFFMASK;
	offset&=IP_OFFMASK;
	offset<<=3; /* 8 byte granularity */

	ihl=iph->ihl<<2;

	len=ntohs(iph->tot_len);

	end = offset + len - ihl;

	if ( (flags&IP_MF)==0 ) {
		if ( end < qp->len ||
			((qp->last_in & LAST_IN) && end!=qp->len)) {
			goto ipcorrupt;
		}
		qp->last_in |= LAST_IN;
		qp->len = end;
	}else{
		end &=~7;
		if ( end > qp->len ) {
			if (qp->last_in & LAST_IN) {
				goto ipcorrupt;
			}
			qp->len=end;
		}
	}

	if ( end==offset ) goto ipcorrupt;

	/* Don't bother wasting any more resources
	 * when we know the packet is oversize (invalid) */
	if ( qp->len > 65535 ) return;

	/* Insert data into fragment chain */
	if ( !(me=malloc(sizeof(struct ipfrag))) ) goto ipfrag_oom;
	ipfrag_mem+=sizeof(struct ipfrag);

	/* Find out where to insert this fragment in the list */
	prev=NULL;
	for(next=qp->fragments; next; next=next->next) {
		if ( next->offset >= offset ) break;
		prev=next;
	}

	/* Check we don't overlap the previous fragment */
	if ( prev ) {
		int i=(prev->offset + prev->len) - offset;

		if ( i>0 ) {
			offset += i;
			if ( end <= offset ) goto ipcorrupt;

			chop=i;
			len-=i;
		}
	}

	/* Make sure we don't overlap next packets */
	while(next && next->offset < end) {
		int i = end - next->offset;

		if ( i < next->len ) {
			next->offset+=i;
			next->len-=i;
			next->data+=i;
			break;
		}else{
			struct ipfrag *free_it=next;
			
			next=next->next;
			if ( prev ){
				prev->next=next;
			}else{
				qp->fragments=next;
				qp->num_frags--;
			}

			qp->meat -= free_it->len;
			free(free_it);
			ipfrag_mem-=sizeof(struct ipfrag);
		}
	}

	/* Make the fragment */
	me->len=len-(chop+ihl);
	me->offset=offset+chop;

	if ( offset ) chop+=ihl;

	/* IP defragmentation can be zerocopy */
	if ( pkt->flags & FP_CLONE ) {
		int alen=me->len;
		
		if ( !offset ) alen+=ihl;
		
		if ( !(me->fdata=malloc(alen)) ) {
			free(me);
			goto ipfrag_oom;
		}
		memcpy(me->fdata, ((void *)iph)+chop, alen);
		me->free=1; 
	}else {
		me->fdata=((void *)iph)+chop;
		me->free=0;
	}

	me->data=me->fdata;

	/* Insert the fragment */
	me->next=next;
	if ( prev ) {
		prev->next=me;
	}else{
		qp->fragments=me;
		qp->num_frags++;
	}

	/* Finish up */
	qp->cap=pkt->capture;
	qp->flags=pkt->flags;
	qp->meat+=me->len;
	if ( !offset ) qp->last_in |= FIRST_IN;

	dprintf("0x%x: got a fragment\n", (unsigned int)qp);

	return;

ipcorrupt:
	dprintf("Corrupt IP fragment\n");
	return;

ipfrag_oom:
	ip_evictor(qp);
	return;
}

/* Preprocessor entry point */
void ipfrag_process(struct packet *pkt, unsigned int i, void *priv)
{
	struct pkt_iphdr *iph;
	unsigned int hash;
	struct ipq *q;

	iph=pkt->layer[i].h.ip;
	if ( iph->frag_off & ipfmask) {
		if ( (q=ip_find(iph, &hash)) ) {
			ipfrag_queue(hash, q, pkt, iph);
			if ( q->last_in == (FIRST_IN|LAST_IN) &&
				q->meat == q->len ) {
				ipfrag_reassemble(q);
			}
		}
	}
}

int ipfrag_init(char *args)
{
	/* TODO: Process mem_hi and mem_lo */
	if (preproc_activate("ip", (proc_preprocess)ipfrag_process, NULL)) {
		printf("ipfrag: IP defragmentation ENABLED.\n");
		return 1;
	}else{
		return 0;
	}
}

int _PLUGIN_INIT (struct plugin_in *in, struct plugin_out *out)
{
	proc_preproc_add preproc_add;
	proc_decode_proto decode_proto;
	int ok=0;
	
	plugin_check(in, out);
	
	PLUGIN_ID("preproc.ipfrag", "IP defragmentation");
	PLUGIN_VERSION(0, 2);
	PLUGIN_AUTHOR("Gianni Tedesco", "gianni@ecsc.co.uk");
	PLUGIN_LICENSE("GPL");
	
	if ( !(preproc_add=in->import("preproc.add")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_activate=in->import("preproc.activate")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(preproc_dispatch=in->import("preproc.dispatch")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	if ( !(decode_proto=in->import("decode.proto")) ) {
		return PLUGIN_ERR_OBJECT;
	}

	ok+=preproc_add("ipfrag", ipfrag_init);
	memset(&ipq_hash, 0, sizeof(ipq_hash));
	
	return (ok) ? PLUGIN_ERR_OK : PLUGIN_ERR_FAIL;
}

int _PLUGIN_UNLOAD (int code) {
	printf("ipfrag: %u reassembled packets, %u errors, %uKB left\n",
		ipfrag_reassembled, ipfrag_errs, ipfrag_mem/1024);
	return PLUGIN_ERR_OK;
}
